/* 
 * tclStringObj.c --
 *
 *	This file contains procedures that implement the Tcl string object
 *	type. These string objects are "dynamic" in that they are grown as
 *	necessary and are efficient to append to. The name "string object"
 *	is slightly confusing because C char*'s are also called
 *	"strings". We try to be consistent in referring to these Tcl objects
 *	as "string objects" and char*'s as "strings".
 *
 * Copyright (c) 1995, 1996 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) tclStringObj.c 1.19 97/01/21 21:35:21
 */

#include "tclInt.h"

/*
 * Prototypes for procedures defined later in this file:
 */

static void		DupStringInternalRep _ANSI_ARGS_((Tcl_Obj *objPtr,
			    Tcl_Obj *copyPtr));
static void		FreeStringInternalRep _ANSI_ARGS_((Tcl_Obj *objPtr));
static int		SetStringFromAny _ANSI_ARGS_((Tcl_Interp *interp,
			    Tcl_Obj *objPtr));
static void		UpdateStringOfString _ANSI_ARGS_((Tcl_Obj *objPtr));

/*
 * The structure below defines the string Tcl object type by means of
 * procedures that can be invoked by generic object code.
 */

Tcl_ObjType tclStringType = {
    "string",				/* name */
    FreeStringInternalRep,		/* freeIntRepProc */
    DupStringInternalRep,		/* dupIntRepProc */
    UpdateStringOfString,		/* updateStringProc */
    SetStringFromAny			/* setFromAnyProc */
};

/*
 *----------------------------------------------------------------------
 *
 * Tcl_NewStringObj --
 *
 *	This procedure is normally called when not debugging: i.e., when
 *	TCL_MEM_DEBUG is not defined. It creates a new string object and
 *	initializes it from the byte pointer and length arguments.
 *
 *	When TCL_MEM_DEBUG is defined, this procedure just returns the
 *	result of calling the debugging version Tcl_DbNewStringObj.
 *
 * Results:
 *	The newly created string object is returned.
 *
 * Side effects:
 *	The new object's internal string representation will be set to a
 *	copy of the length bytes starting at "bytes". If "length" is
 *	negative, use bytes up to the first NULL byte; i.e., assume "bytes"
 *	points to a C-style NULL-terminated string. The object's type is set
 *	to NULL. An extra NULL is added to the end of the new object's byte
 *	array.
 *
 *----------------------------------------------------------------------
 */

#ifdef TCL_MEM_DEBUG
#undef Tcl_NewStringObj

Tcl_Obj *
Tcl_NewStringObj(bytes, length)
    register char *bytes;	/* Points to the first of the length bytes
				 * used to initialize the new object. */
    register int length;	/* The number of bytes to copy from "bytes"
				 * when initializing the new object. If 
				 * negative, use bytes up to the first
				 * NULL byte. */
{
    return Tcl_DbNewStringObj(bytes, length, "unknown", 0);
}

#else /* if not TCL_MEM_DEBUG */

Tcl_Obj *
Tcl_NewStringObj(bytes, length)
    register char *bytes;	/* Points to the first of the length bytes
				 * used to initialize the new object. */
    register int length;	/* The number of bytes to copy from "bytes"
				 * when initializing the new object. If 
				 * negative, use bytes up to the first
				 * NULL byte. */
{
    register Tcl_Obj *objPtr;

    if (length < 0) {
	length = strlen(bytes);
    }
    TclNewObj(objPtr);
    TclInitStringRep(objPtr, bytes, length);
    return objPtr;
}
#endif /* TCL_MEM_DEBUG */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_DbNewStringObj --
 *
 *	This procedure is normally called when debugging: i.e., when
 *	TCL_MEM_DEBUG is defined. It creates new string objects. It is the
 *	same as the Tcl_NewStringObj procedure above except that it calls
 *	Tcl_DbCkalloc directly with the file name and line number from its
 *	caller. This simplifies debugging since then the checkmem command
 *	will report the correct file name and line number when reporting
 *	objects that haven't been freed.
 *
 *	When TCL_MEM_DEBUG is not defined, this procedure just returns the
 *	result of calling Tcl_NewStringObj.
 *
 * Results:
 *	The newly created string object is returned.
 *
 * Side effects:
 *	The new object's internal string representation will be set to a
 *	copy of the length bytes starting at "bytes". If "length" is
 *	negative, use bytes up to the first NULL byte; i.e., assume "bytes"
 *	points to a C-style NULL-terminated string. The object's type is set
 *	to NULL. An extra NULL is added to the end of the new object's byte
 *	array.
 *
 *----------------------------------------------------------------------
 */

#ifdef TCL_MEM_DEBUG

Tcl_Obj *
Tcl_DbNewStringObj(bytes, length, file, line)
    register char *bytes;	/* Points to the first of the length bytes
				 * used to initialize the new object. */
    register int length;	/* The number of bytes to copy from "bytes"
				 * when initializing the new object. If 
				 * negative, use bytes up to the first
				 * NULL byte. */
    char *file;			/* The name of the source file calling this
				 * procedure; used for debugging. */
    int line;			/* Line number in the source file; used
				 * for debugging. */
{
    register Tcl_Obj *objPtr;

    if (length < 0) {
	length = strlen(bytes);
    }
    TclDbNewObj(objPtr, file, line);
    TclInitStringRep(objPtr, bytes, length);
    return objPtr;
}

#else /* if not TCL_MEM_DEBUG */

Tcl_Obj *
Tcl_DbNewStringObj(bytes, length, file, line)
    register char *bytes;	/* Points to the first of the length bytes
				 * used to initialize the new object. */
    register int length;	/* The number of bytes to copy from "bytes"
				 * when initializing the new object. If 
				 * negative, use bytes up to the first
				 * NULL byte. */
    char *file;			/* The name of the source file calling this
				 * procedure; used for debugging. */
    int line;			/* Line number in the source file; used
				 * for debugging. */
{
    return Tcl_NewStringObj(bytes, length);
}
#endif /* TCL_MEM_DEBUG */

/*
 *----------------------------------------------------------------------
 *
 * Tcl_SetStringObj --
 *
 *	Modify an object to hold a string that is a copy of the bytes
 *	indicated by the byte pointer and length arguments. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The object's string representation will be set to a copy of
 *	the "length" bytes starting at "bytes". If "length" is negative, use
 *	bytes up to the first NULL byte; i.e., assume "bytes" points to a
 *	C-style NULL-terminated string. The object's old string and internal
 *	representations are freed and the object's type is set NULL. We
 *	delay actually setting the object's type to &tclStringType until
 *	necessary, e.g., when Tcl_StringObjAppend is called on the object.
 *	Note that all tclStringType objects have a non-NULL DString
 *	internal representation.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_SetStringObj(objPtr, bytes, length)
    register Tcl_Obj *objPtr;	/* Object whose internal rep to init. */
    char *bytes;		/* Points to the first of the length bytes
				 * used to initialize the object. */
    register int length;	/* The number of bytes to copy from "bytes"
				 * when initializing the object. If 
				 * negative, use bytes up to the first
				 * NULL byte.*/
{
    register Tcl_ObjType *oldTypePtr = objPtr->typePtr;

    /*
     * Free any old string rep, then set the string rep to a copy of
     * the length bytes starting at "bytes".
     */

    if (objPtr->bytes != NULL) {
	ckfree((char *) objPtr->bytes);
    }

    if (length < 0) {
	length = strlen(bytes);
    }
    TclInitStringRep(objPtr, bytes, length);
        
    /*
     * Set the type to NULL and free any internal rep for the old type.
     */

    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }
    objPtr->typePtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_StringObjAppend --
 *
 *	This procedure appends a sequence of bytes to a string object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	"length" bytes starting at "bytes" (or all bytes up to a NULL byte
 *	if length is less than zero) are added to the current value of the
 *	string object's internal DString representation. If necessary,
 *	objPtr will be converted to a string object and an internal DString
 *	representation will be created. The string representation will be
 *	set NULL after freeing any old value. Memory for the DString
 *	representation will be reallocated if needed to accomodate the
 *	string's new size.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_StringObjAppend(objPtr, bytes, length)
    register Tcl_Obj *objPtr;	/* Points to the object to append to. */
    char *bytes;		/* Points to the first of the length bytes
				 * to append to the string object. */
    register int length;	/* The number of bytes to append from
				 * "bytes". If < 0, then append all bytes
				 * up to NULL byte. */
{
    register Tcl_DString *dsPtr;

    if (objPtr->typePtr != &tclStringType) {
	if (SetStringFromAny((Tcl_Interp *) NULL, objPtr) != TCL_OK) {
	    panic("Tcl_StringObjAppend could not convert to string");
	}
    }
    dsPtr = (Tcl_DString *) objPtr->internalRep.otherValuePtr;

    /*
     * Use Tcl_DStringAppend to add the "length" bytes to the end of
     * the string. Note that it calls strlen if "length" is negative
     * and that it always adds a NULL at the end.
     */
    
    Tcl_DStringAppend(dsPtr, bytes, length);

    /*
     * Set the string representation NULL since it is no longer valid.
     */

    if (objPtr->bytes != NULL) {
	ckfree(objPtr->bytes);
	objPtr->bytes = NULL;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Tcl_StringObjAppendObj --
 *
 *	This procedure appends an object's string representation to a
 *	string object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The bytes of srcPtr's string representation are added to the current
 *	value of the objPtr's internal DString representation. If necessary,
 *	objPtr will be converted to a string object and an internal DString
 *	representation will be created. The string representation will be
 *	set NULL after freeing any old value. Memory for the DString
 *	representation will be reallocated if needed to accomodate the
 *	string's new size. srcPtr's string representation will be
 *	regenerated if necessary. If srcPtr does not have a non-NULL string
 *	rep, no bytes will be added to objPtr.
 *
 *----------------------------------------------------------------------
 */

void
Tcl_StringObjAppendObj(objPtr, srcPtr)
    register Tcl_Obj *objPtr;	/* Points to the string obj to append to. */
    register Tcl_Obj *srcPtr;	/* Points to the object whose string rep
				 * will be appended to objPtr. */
{
    register Tcl_DString *dsPtr;
    char *bytes;
    int length;

    if (objPtr->typePtr != &tclStringType) {
	if (SetStringFromAny((Tcl_Interp *) NULL, objPtr) != TCL_OK) {
	    panic("Tcl_StringObjAppendObj could not convert to string");
	}
    }

    dsPtr = (Tcl_DString *) objPtr->internalRep.otherValuePtr;
    bytes = TclGetStringFromObj(srcPtr, &length);
    Tcl_DStringAppend(dsPtr, bytes, length); /* adds NULL after end */

    /*
     * Set the string representation NULL since it is no longer valid.
     */

    if (objPtr->bytes != NULL) {
	ckfree(objPtr->bytes);
	objPtr->bytes = NULL;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * FreeStringInternalRep --
 *
 *	Deallocate any storage associated with a string object's internal
 *	representation.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Frees the string object's dynamic string.
 *
 *----------------------------------------------------------------------
 */

static void
FreeStringInternalRep(objPtr)
    register Tcl_Obj *objPtr;	/* Object whose internal rep to free. */
{
    register Tcl_DString *dsPtr =
	(Tcl_DString *) objPtr->internalRep.otherValuePtr;

    Tcl_DStringFree(dsPtr);
    ckfree((char *) dsPtr);
    objPtr->internalRep.otherValuePtr = NULL;
}

/*
 *----------------------------------------------------------------------
 *
 * DupStringInternalRep --
 *
 *	Initialize the internal representation of a new Tcl_Obj to a
 *	copy of the internal representation of an existing string object,
 *	thus making the new Tcl_Obj into a string object.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	srcPtr's DString internal representation pointer must not be NULL.
 *	copyPtr's internal rep is set to a copy of srcPtr's DString internal
 *	representation.
 *
 *----------------------------------------------------------------------
 */

static void
DupStringInternalRep(srcPtr, copyPtr)
    Tcl_Obj *srcPtr;		/* Object with internal rep to copy. */
    register Tcl_Obj *copyPtr;	/* Object with internal rep to set. */
{
    Tcl_DString *srcDsPtr = (Tcl_DString *) srcPtr->internalRep.otherValuePtr;
    Tcl_DString *copyDsPtr;

    copyDsPtr = (Tcl_DString *) ckalloc(sizeof(Tcl_DString));
    Tcl_DStringInit(copyDsPtr);
    Tcl_DStringAppend(copyDsPtr, srcDsPtr->string, srcDsPtr->length);
    copyPtr->internalRep.otherValuePtr = (VOID *) copyDsPtr;
    copyPtr->typePtr = &tclStringType;
}

/*
 *----------------------------------------------------------------------
 *
 * SetStringFromAny --
 *
 *	Generate the internal form for the Tcl string object "objPtr".
 *
 * Results:
 *	This operation always succeeds and returns TCL_OK.
 *
 * Side effects:
 *	The type of "objPtr" is set to &tclStringType and the internal rep
 *	DString pointer is set to point to a newly allocated DString
 *	containing the object's string rep bytes.
 *
 *----------------------------------------------------------------------
 */

static int
SetStringFromAny(interp, objPtr)
    Tcl_Interp *interp;		/* Used for error reporting if not NULL. */
    register Tcl_Obj *objPtr;	/* The object to convert. */
{
    register Tcl_ObjType *oldTypePtr = objPtr->typePtr;
    register Tcl_DString *dsPtr;

    /*
     * Make sure "objPtr"s string representation is up-to-date. Note that
     * the string rep's byte pointer might still be NULL afterwards.
     */

    if (objPtr->bytes == NULL) {
	if (objPtr->typePtr != NULL) {
	    objPtr->typePtr->updateStringProc(objPtr);
	}
    }
    
    /*
     * Free any internal rep for the old type. Create the DString internal
     * rep for the string object. Note that this code works even if there
     * are embedded NULLs.
     */

    if ((oldTypePtr != NULL) && (oldTypePtr->freeIntRepProc != NULL)) {
	oldTypePtr->freeIntRepProc(objPtr);
    }

    dsPtr = (Tcl_DString *) ckalloc(sizeof(Tcl_DString));
    Tcl_DStringInit(dsPtr);
    if (objPtr->bytes != NULL) {
	Tcl_DStringAppend(dsPtr, objPtr->bytes, objPtr->length);
    }
    objPtr->internalRep.otherValuePtr = (VOID*) dsPtr;
    objPtr->typePtr = &tclStringType;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * UpdateStringOfString --
 *
 *	Update the string representation for a Tcl string object.
 *	Note: This procedure does not free an existing old string rep
 *	so storage will be lost if this has not already been done. 
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	The object's string rep is set from the internal DString rep.
 *	The DString internal representation pointer must not be NULL.
 *
 *----------------------------------------------------------------------
 */

static void
UpdateStringOfString(objPtr)
    register Tcl_Obj *objPtr;	/* Object with string rep to update. */
{
    register Tcl_DString *dsPtr = (Tcl_DString *)
	    objPtr->internalRep.otherValuePtr;
    register int length = dsPtr->length;

    objPtr->bytes = (char *) ckalloc((unsigned) (length + 1));
    memcpy((VOID *) objPtr->bytes, (VOID *) dsPtr->string, (size_t) length);
    objPtr->bytes[length] = '\0';
    objPtr->length = length;
}
